﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;

namespace AAL77FDRDecompressor
{
    //Version 1.0

    //This class allows the input file to be read in units of 128 byte pages.
    public class PageFileStream
    {
        private const int blockLength = 0x20000;

        private FileStream inputFile = null;
        private int inputFileOffset = 0;
        private int inputFileStart = 0;
        private int inputFileLength = 0;
        private int inputFilePos = 0;

        public PageFileStream(string fileName)
        {
            inputFile = new FileStream(fileName, FileMode.Open, FileAccess.Read);
            int fl = (int)inputFile.Length;
            bool finished = false;
            //Search through each pair of 64KB blocks for the earliest recorded data 
            //and find out how long the data is not including the text header at the 
            //beginning of the file.
            for (int fp = fl - blockLength; (fp >= 0) && (!finished); fp -= blockLength)
            {
                inputFile.Seek(fp, SeekOrigin.Begin);
                byte[] headerBytes = new byte[2];
                inputFile.Read(headerBytes, 0, headerBytes.Length);
                int headerWord = (headerBytes[1] << 8) | headerBytes[0];
                switch (headerWord)
                {
                    case 0xFE6B:
                        //This is a pair of 64KB blocks containing valid data.
                        inputFileOffset = fp;
                        break;
                    case 0xFFFF:
                        //This is the pair of blank blocks which is left erased and ready to 
                        //be used when the current pair of blocks is full so the earliest 
                        //recorded data is the next pair of 64KB blocks.
                        inputFileOffset = fp;
                        inputFileStart = fp + blockLength;
                        break;
                    default:
                        //This is invalid data so we must be looking through the text header 
                        //at the beginning of the file.
                        finished = true;
                        break;
                }
            }
            if (inputFileStart >= fl)
                inputFileStart = inputFileOffset;
            //Calculate the start point and length of the data in units of 128 byte pages.
            inputFileStart = (inputFileStart - inputFileOffset) >> 7;
            inputFileLength = (fl - inputFileOffset) >> 7;
            Seek(0);
        }

        public int Length
        {
            get
            {
                return inputFileLength - (blockLength >> 7);
            }
        }

        public int BytePos(int fp)
        {
            //Calculate the byte file offset of a file offset measured in units of 128 byte 
            //pages. Offset 0 is the earliest recorded data. The recorded data wraps around 
            //to the beginning of the file as in a standard ring buffer.
            int pagePos = fp + inputFileStart;
            if (pagePos >= inputFileLength)
                pagePos -= inputFileLength;
            return ((pagePos << 7) + inputFileOffset);
        }

        public void Seek(int fp)
        {
            //Seek in to the file in units of 128 byte pages.
            inputFile.Seek(BytePos(fp), SeekOrigin.Begin);
            inputFilePos = fp;
        }

        public byte[] ReadPage()
        {
            //Read a single 128 byte page.
            byte[] page = new byte[128];
            inputFile.Read(page, 0, page.Length);
            if ((++inputFilePos + inputFileStart) == inputFileLength)
                Seek(inputFilePos);
            return page;
        }

        public void Close()
        {
            inputFile.Close();
        }
    }

    abstract class HuffmanCodeNode
    {
        public bool IsLeaf
        {
            get
            {
                return this is HuffmanCodeLeafNode;
            }
        }
    }

    //This class is for interior (non leaf) nodes of the Huffman tree.
    class HuffmanCodeInteriorNode : HuffmanCodeNode
    {
        private HuffmanCodeNode[] child = null;

        public HuffmanCodeInteriorNode()
        {
            child = new HuffmanCodeNode[2];
            for (int i = 0; i < child.Length; i++)
                child[i] = null;
        }

        public HuffmanCodeNode GetChild(int i)
        {
            return child[i];
        }

        //Add Huffman codes and values (referred to as differences here) to the Huffman tree.
        //The Huffman codes have the form pppppxxxxx in binary.
        //The p bits are from prefix which is a string representation of a binary number. 
        //The x bits are related to the differences. There are a total of numBits x bits.
        //A non-negative differences is considered as a 12 bit signed integer, the first x bit 
        //    is the sign of the integer. The other x bits are a binary representation of 
        //    the absolute value of the integer. 
        //A difference of -1 represents a loss of synchronisation and the end of a segment of 
        //    compressed data.
        public void AddHuffmanCodesForPrefix(string prefix, int startDifference, int numBits)
        {
            //Add the prefix to the Huffman tree.
            HuffmanCodeNode huffmanCodeNode = this;
            for (int i = 0; i < prefix.Length; i++)
            {
                int bit = int.Parse(prefix[i].ToString());
                HuffmanCodeNode nextHuffmanCodeNode = ((HuffmanCodeInteriorNode)huffmanCodeNode).child[bit];
                if (nextHuffmanCodeNode == null)
                {
                    if ((numBits == 0) && (i == prefix.Length - 1))
                    {
                        ((HuffmanCodeInteriorNode)huffmanCodeNode).child[bit] = new HuffmanCodeLeafNode(startDifference);
                        return;
                    }
                    else
                    {
                        nextHuffmanCodeNode = new HuffmanCodeInteriorNode();
                        ((HuffmanCodeInteriorNode)huffmanCodeNode).child[bit] = nextHuffmanCodeNode;
                    }
                }
                huffmanCodeNode = nextHuffmanCodeNode;
            }
            //Add each difference to the Huffman tree, starting with startDifference.
            Stack<HuffmanCodeNode> huffmanCodeNodeStack = new Stack<HuffmanCodeNode>();
            int difference = startDifference;
            while (true)
            {
                bool createdHuffmanCodeNode = false;
                for (int i = 0; i < ((HuffmanCodeInteriorNode)huffmanCodeNode).child.Length; i++)
                {
                    HuffmanCodeNode nextHuffmanCodeNode = ((HuffmanCodeInteriorNode)huffmanCodeNode).child[i];
                    if (nextHuffmanCodeNode == null)
                    {
                        if (huffmanCodeNodeStack.Count < numBits - 1)
                        {
                            nextHuffmanCodeNode = new HuffmanCodeInteriorNode();
                            ((HuffmanCodeInteriorNode)huffmanCodeNode).child[i] = nextHuffmanCodeNode;
                            huffmanCodeNodeStack.Push(huffmanCodeNode);
                            huffmanCodeNode = nextHuffmanCodeNode;
                        }
                        else
                        {
                            ((HuffmanCodeInteriorNode)huffmanCodeNode).child[i] = new HuffmanCodeLeafNode(difference);
                            if (difference < 0x800)
                                difference++;
                            else
                                difference--;
                        }
                        createdHuffmanCodeNode = true;
                        break;
                    }
                }
                if (!createdHuffmanCodeNode)
                {
                    if (huffmanCodeNodeStack.Count < 1)
                        break;
                    huffmanCodeNode = huffmanCodeNodeStack.Pop();
                    if (huffmanCodeNodeStack.Count < 1)
                        difference = 0x1000 - startDifference;
                }
            }
        }
    }

    class HuffmanCodeLeafNode : HuffmanCodeNode
    {
        private int difference = 0;

        public HuffmanCodeLeafNode(int aDifference)
        {
            difference = aDifference;
        }

        public int Difference
        {
            get
            {
                return difference;
            }
        }
    }


    //The following class holds the current state of one of the two independent flight data streams.

    //Information about the format in which the FDR records the data can be found in the post 
    //at http://forums.randi.org/showpost.php?p=2022420&postcount=85 and in figure 4.2-2 of 
    //the document http://www.asc.gov.tw/author_files/ASC-TRT-99-007.pdf

    //Flight Data Stream 0 consists of the first, third, fifth and so on blocks of 512 pages from 
    //the PageFileStream.

    //Flight Data Stream 1 consists of the second, fourth, sixth and so on blocks of 512 pages from 
    //the PageFileStream.

    //The first 2 pages from each block of 512 pages are header pages that also contain Bad Page 
    //Maps. This program ignores these pages.

    //The words within the other pages in the figure are 16 bit words stored Least Significant Byte 
    //first and the Hamming code & page parity occupy the last 11 bits of each page. Within each 
    //16 bit word, the most significant bit is the first bit in the bit stream. Some pages are 
    //blank consisting entirely of 0xFFFF words and need to be skipped. The Hamming code and page 
    //parity are ignored by this program. My AAL77 Hamming code and page parity checker program 
    //does read them however.

    //Data is recorded by the FDR in frames separated by the 16 bit frame marker word 0xFEFE. A 
    //frame consists of 4 subframes of 256 12 bit words. A subframe consists of 1 second of data.
    //The FDR does not record the first 12 bit word of each subframe since they are sync words 
    //and are always the same. Therefore a complete frame is stored as 1020 * 12 bit words. The 
    //subframes are numbered 1 to 4 and appear in that order in each frame.

    //The frames are recorded in a compressed form using Huffman coding. At the beginning of each 
    //section of compressed data within a flight data stream and periodically every 75 frames, a
    //baseline frame is recorded that consists of the Huffman codes of each of the 1020 * 12 bit 
    //words. Each baseline frame is preceded with the bits 1000 and a 16 bit frame counter.

    //The other frames are recorded by calculating the difference between each of the 1020 * 12 bit
    //words in this frame and the one before it in this flight data stream. The Huffman codes of 
    //these differences are recorded for those frames. Each of these frames are preceded with a 
    // 0 bit.

    //Except for the baseline frame at the beginning of each section of compressed data in a flight
    //data stream, the baseline frames are additional copies of the current frame, i.e. the 
    //baseline frames need to not appear in the decoded output, to avoid 2 copies of the same frame
    //of data appearing. This program does not output the first baseline frame at the beginning of 
    //each section of compressed data, so that adds up to 8 seconds of data at the beginning of 
    //each section of compressed data that is not decoded.

    //Frames are alternately stored in each of the flight data streams so one flight data stream 
    //will consist of the first, third, fifth and so on frames of data and the other flight data
    //stream will have the second, fourth, sixth and so on frames of data. A frame counter is 
    //recorded along with each baseline frame that makes it possible to determine which flight data
    //stream contains the first frame of data.

    //A special Huffman code is used to mark the loss of synchronisation and the subsequent end of 
    //the section of compressed data. This can occur at any point within the frame which is why the
    //last frame of data within a section of compressed data is less than 1020 * 12 bit words.

    //In between the sections of compressed data, up to 4 subframes of uncompressed data is 
    //generated in the file each time the FDR is started up which I believe happens when the 
    //engines are started. My AAL77 FDR partial decoder program decodes the uncompressed data.

    class FlightDataStream
    {
        enum FrameStates { StartState, GetFrameType, GetCounter, DecompressFrame, WaitUntilEnd };

        private int[] frame = new int[1020];
        private byte[] page = null;
        private HuffmanCodeNode huffmanCodeTree = null;
        private FrameStates frameState = FrameStates.StartState;
        private int numFlightDataStreams = 0;
        private int subFrameNumber = 0;
        private int superFrameNumber = 0;
        private int counter = -1;
        private int lastFramePos = 0;
        private int bitPos = 0;
        private int pageNum = 0;
        private int filePos = 0;
        private int lastWordPos = -1;
        private int w = 0;
        private bool syncLost = false;
        private bool endOfData = false;
        private bool baseLineFrame = false;
        private bool badFrame = false;

        public FlightDataStream(HuffmanCodeNode aHuffmanCodeTree, int flightDataStreamNumber, int aNumFlightDataStreams)
        {
            huffmanCodeTree = aHuffmanCodeTree;
            numFlightDataStreams = aNumFlightDataStreams;
            for (int i = 0; i < frame.Length; i++)
                frame[i] = 0;
            pageNum = (flightDataStreamNumber << 9) + 2;
        }

        public int Counter
        {
            get
            {
                return counter;
            }
        }

        public bool EndOfProcessedFrames
        {
            get
            {
                return (endOfData && (frameState == FrameStates.StartState));
            }
        }

        public void ReadFrame()
        {
            HuffmanCodeNode huffmanCodeNode = huffmanCodeTree;
            int framePos = 0;
            int lastFramePosWith0Bit = -1;
            int frameType = 0;
            int frameTypeBits = 0;
            int counterBits = 0;
            syncLost = false;
            while (!endOfData)
            {
                //The last 11 bits of the 128 byte (1024 bit) page are the Hamming Code and page
                //parity which this code skips.
                if ((bitPos >= 1013) || (page == null))
                {
                    bitPos = 0;
                    bool emptyPage = true;
                    //Skip empty pages (pages where all bits are 1)
                    while (emptyPage)
                    {
                        Program.inputFile.Seek(pageNum);
                        page = Program.inputFile.ReadPage();
                        pageNum++;
                        if ((pageNum & 0x1FF) == 0)
                            //We are at the end of a block of pages so go the first data page 
                            //within the next block of pages in this flight data stream.
                            pageNum += 0x202;
                        for (int bp = 0; bp < page.Length; bp++)
                            if (page[bp] != 0xFF)
                            {
                                emptyPage = false;
                                break;
                            }
                    }
                }
                if (pageNum >= Program.inputFile.Length)
                {
                    endOfData = true;
                    filePos = Program.inputFile.BytePos(pageNum);
                    break;
                }
                int wordPos = bitPos >> 4;
                if (wordPos != lastWordPos)
                {
                    w = (page[(wordPos << 1) | 1] << 8) | page[(wordPos << 1)];
                    lastWordPos = wordPos;
                    if ((w == 0xFEFE) && (wordPos != 63) && (frameState == FrameStates.WaitUntilEnd))
                    {
                        bitPos += 0x10;
                        filePos = Program.inputFile.BytePos(pageNum - 1) + ((bitPos >> 3) & 0x7E);
                        //We have now finished reading this frame
                        break;
                    }
                }
                int bit = (w >> (0xF - (bitPos++ & 0xF))) & 1;
                switch (frameState)
                {
                    case FrameStates.StartState:
                        huffmanCodeNode = huffmanCodeTree;
                        framePos = 0;
                        lastFramePosWith0Bit = -1;
                        badFrame = false;
                        if (baseLineFrame = (bit == 1))
                        {
                            //This could be a baseline frame, read the frame type.
                            frameType = 0;
                            frameTypeBits = 0;
                            frameState = FrameStates.GetFrameType;
                        }
                        else
                        {
                            //Start decoding this non baseline frame.
                            counter += numFlightDataStreams;
                            frameState = FrameStates.DecompressFrame;
                        }
                        break;
                    case FrameStates.GetFrameType:
                        frameType = (frameType << 1) | bit;
                        if (++frameTypeBits >= 3)
                            if (frameType == 0)
                            {
                                //Read the counter value for this baseline frame.
                                counter = 0;
                                counterBits = 0;
                                frameState = FrameStates.GetCounter;
                            }
                            else
                            {
                                //This data is not a compressed frame.
                                badFrame = true;
                                frameState = FrameStates.WaitUntilEnd;
                            }
                        break;
                    case FrameStates.GetCounter:
                        counter = (counter << 1) | bit;
                        if (++counterBits >= 16)
                            //Start decoding this baseline frame.
                            frameState = FrameStates.DecompressFrame;
                        break;
                    case FrameStates.DecompressFrame:
                        if (bit == 0)
                            lastFramePosWith0Bit = framePos;
                        //Follow the Huffman code tree.
                        huffmanCodeNode = ((HuffmanCodeInteriorNode)huffmanCodeNode).GetChild(bit);
                        if (huffmanCodeNode.IsLeaf)
                        {
                            //We now have read a complete Huffman code. Get the difference.
                            int difference = ((HuffmanCodeLeafNode)huffmanCodeNode).Difference;
                            if (difference < 0)
                            {
                                //The Huffman code just read was the special loss of 
                                //synchronisation code. Wait for the frame marker word.
                                syncLost = true;
                                badFrame = true;
                                frameState = FrameStates.WaitUntilEnd;
                            }
                            else
                            {
                                //Add the difference to the 12 bit word for a non-baseline frame, 
                                //or overwrite the 12 bit word with the difference for a baseline 
                                //frame.
                                frame[framePos] = ((baseLineFrame ? 0 : frame[framePos]) + difference) & 0xFFF;
                                //Start reading the next Huffman code.
                                huffmanCodeNode = huffmanCodeTree;
                                if (++framePos >= frame.Length)
                                    //We now have a full frame, wait for the frame marker word.
                                    frameState = FrameStates.WaitUntilEnd;
                            }
                        }
                        break;
                    case FrameStates.WaitUntilEnd:
                        //The rest of the bits remaining before the frame marker word should all be 1
                        if (bit != 1)
                            badFrame = true;
                        break;
                }
            }
            lastFramePos = endOfData ? lastFramePosWith0Bit + 1 : framePos;
            if (lastFramePos > 764)
                superFrameNumber = (frame[764] >> 8) + 1;
            else
                if (superFrameNumber >= 0)
                    superFrameNumber = ((superFrameNumber + numFlightDataStreams - 1) % 16) + 1;
        }

        public void ReadGoodBaseLineFrame()
        {
            superFrameNumber = -1;
            //Keep reading frames until we get a good baseline frame.
            while (((!baseLineFrame) || badFrame) && (!endOfData))
            {
                frameState = FrameStates.StartState;
                ReadFrame();
            }
        }

        string BinaryStringOfWord(int word)
        {
            string result = "";
            for (int i = 11; i >= 0; i--)
                result += ((word >> i) & 1).ToString();
            return result;
        }

        //This method writes the frame contents to the output file.
        public bool RecordFrame()
        {
            int[] syncWords = { 583, 1464, 2631, 3512 };

            //Only record non baseline frames that have some data in them.
            if ((!baseLineFrame) && (lastFramePos > 0))
            {
                int framePos = 0;
                //Cycle through each subframe within the frame.
                for (subFrameNumber = 1; subFrameNumber <= 4; subFrameNumber++)
                {
                    if (lastFramePos <= framePos)
                        break;
                    string line = BinaryStringOfWord(syncWords[subFrameNumber - 1]);
                    for (int wordNum = 2; wordNum <= 256; wordNum++)
                        if (lastFramePos > framePos)
                            line += " " + BinaryStringOfWord(frame[framePos++]);
                        else
                            break;
                    Program.outputFile.WriteLine(line);
                }
                if (syncLost)
                    Program.outputFile.WriteLine("Sync Lost");
            }
            frameState = FrameStates.StartState;
            return syncLost;
        }
    }

    class GenerateOutput
    {
        public bool Start()
        {
            //Construct the Huffman code tree.
            HuffmanCodeInteriorNode huffmanCodeTree = new HuffmanCodeInteriorNode();
            huffmanCodeTree.AddHuffmanCodesForPrefix("0", 0, 0);
            huffmanCodeTree.AddHuffmanCodesForPrefix("10", 1, 2);
            huffmanCodeTree.AddHuffmanCodesForPrefix("110", 3, 3);
            huffmanCodeTree.AddHuffmanCodesForPrefix("1110", 7, 4);
            huffmanCodeTree.AddHuffmanCodesForPrefix("111100", 15, 5);
            huffmanCodeTree.AddHuffmanCodesForPrefix("1111100", 31, 6);
            huffmanCodeTree.AddHuffmanCodesForPrefix("1111010", 63, 7);
            huffmanCodeTree.AddHuffmanCodesForPrefix("111111", 127, 8);
            huffmanCodeTree.AddHuffmanCodesForPrefix("1111101", 255, 9);
            huffmanCodeTree.AddHuffmanCodesForPrefix("11110110", 511, 10);
            huffmanCodeTree.AddHuffmanCodesForPrefix("111101110", 1023, 11);
            huffmanCodeTree.AddHuffmanCodesForPrefix("1111011110", 2047, 2);
            huffmanCodeTree.AddHuffmanCodesForPrefix("1111011111", -1, 0);  //This is the special Huffman code that marks loss of synchronisation.
            FlightDataStream[] flightDataStream = new FlightDataStream[2];
            for (int flightDataStreamNumber = 0; flightDataStreamNumber < flightDataStream.Length; flightDataStreamNumber++)
                flightDataStream[flightDataStreamNumber] = new FlightDataStream(huffmanCodeTree, flightDataStreamNumber, flightDataStream.Length);
            bool syncLost = true;
            while (true)
            {
                if (syncLost)
                {
                    //We have reached the end of a section of compressed data. Find the next good
                    //baseline frame for both flight data streams.
                    for (int flightDataStreamNumber = 0; flightDataStreamNumber < flightDataStream.Length; flightDataStreamNumber++)
                        flightDataStream[flightDataStreamNumber].ReadGoodBaseLineFrame();
                    syncLost = false;
                }
                //Determine which flight data stream has the current frame with the lowest counter
                //value. This will be the frame to be recorded to the output next.
                int firstFlightDataStreamNumber = -1;
                int firstCounter = -1;
                for (int flightDataStreamNumber = 0; flightDataStreamNumber < flightDataStream.Length; flightDataStreamNumber++)
                    if (!flightDataStream[flightDataStreamNumber].EndOfProcessedFrames)
                    {
                        int counter = flightDataStream[flightDataStreamNumber].Counter;
                        if ((firstCounter < 0) || (((counter - firstCounter) & 0x8000) != 0))
                        {
                            firstFlightDataStreamNumber = flightDataStreamNumber;
                            firstCounter = counter;
                        }
                    }
                if (firstFlightDataStreamNumber < 0)
                    //No data to be recorded was found so the output is finished.
                    break;
                if (flightDataStream[firstFlightDataStreamNumber].RecordFrame())
                    //The loss of synchronisation code was found within the frame just recorded to 
                    //the output so we are at the end of this section of compressed data.
                    syncLost = true;
                flightDataStream[firstFlightDataStreamNumber].ReadFrame();
            }
            return (true);
        }
    }

    class Program
    {
        static public PageFileStream inputFile = null;
        static public StreamWriter outputFile = null;

        static void Main(string[] args)
        {
            inputFile = new PageFileStream(@"C:\Documents and Settings\Admin\My Documents\Pilots For 911 Truth\090324_0813\AA77 FDR\American 77.fdr");
            outputFile = new StreamWriter(@"C:\Documents and Settings\Admin\My Documents\Pilots For 911 Truth\AAL77 Decompressed.txt");
            GenerateOutput generateOutput = new GenerateOutput();
            generateOutput.Start();
            outputFile.Close();
            inputFile.Close();
        }
    }
}
